BeginPackage["HEPMath`FeynArts`",
  {"HEPMath`Tensors`",
   "HEPMath`Lorentz`",
   "HEPMath`Color`",
   "HEPMath`PaVe`"}]


(* load FeynArts and quarantine clashing symbols *)

Needs["HEPMath`Symbolics`"]

HEPMath`FeynArts`Private`oldContextPath = $ContextPath

$ContextPath=DeleteCases[$ContextPath,
   "HEPMath`Lorentz`"|"HEPMath`Tensors`"|"HEPMath`PaVe`"|
   "HEPMath`FeynArts`"]

Quiet[Needs["FeynArts`"], General::shdw]

Quiet[
  Rename[Global`DiracSpinor, HEPMath`FeynArts`Model`DiracSpinor,
         UpdateContexts->{"FeynArts`*", "HEPMath`FeynArts`Model`*"}];
  Rename[Global`PolarizationVector, HEPMath`FeynArts`Model`PolarizationVector,
         UpdateContexts->{"FeynArts`*", "HEPMath`FeynArts`Model`*"}];
  Rename[Global`Lorentz, HEPMath`FeynArts`Conflict`Lorentz,
         UpdateContexts->{"FeynArts`*", "HEPMath`FeynArts`Conflict`*"}];
  Rename[FeynArts`Indices, HEPMath`FeynArts`Conflict`Indices,
         UpdateContexts->{"FeynArts`*", "HEPMath`FeynArts`Conflict`*"}];
  Rename[FeynArts`CC, HEPMath`FeynArts`Conflict`CC,
         UpdateContexts->{"FeynArts`*", "HEPMath`FeynArts`Conflict`*"}];
  Rename[FeynArts`Initialize`ReadModelFile,
         HEPMath`FeynArts`Private`oldReadModelFile,
         Keep->True];
  ,
  {RuleDelayed::rhs}]

$ContextPath = Join[{"FeynArts`", "HEPMath`FeynArts`Model`"},
                    HEPMath`FeynArts`Private`oldContextPath]


(* additional interface *)

ConvertFeynAmp::usage = "ConvertFeynAmp[amp] converts an expression returned
by the CreateFeynAmp function of the FeynArts package to an expression which
can be manipulated with HEPMath. The returned expression is of the form
 {extrul, amp} where extrul is a list of rules specifying
the indices and masses of the external particles and amp is a sum of
InsertionSum expressions."

ConvertFeynAmp::upol = "Cannot determine type of polarization vector {`1`, `2`}."

InsertionSum::usage = "InsertionSum[expr, subexlist -> 
Insertions[lvl][ins1, ins2, ...]] represents a sum of multiple copies of expr,
where in each copy the subexpressions in subexlist are replaced by
corresponding expressions in ins1, ins2 etc. The insertions ins1, ins2 etc. may
also be of the form list -> Insertions[lvl2][ins3, ins4, ...]."

Polarization::usage = "Polarization[Incoming, n] and Polarization[Outgoing, n]
is the spin polarization index of the n-th incoming our outgoing external
particle, respectively."

RenameInsertions::usage = "RenameInsertions[expr] replaces all inserted
subexpressions in InsertionSum expressions found in expr by unique symbols."

ApplyInsertions::usage = "ApplyInsertions[expr] replaces InsertionSum
expressions in expr by ordinary sums."

Coup::usage = "Coup[sym] denotes a generic coupling in an InsertionSum
expression. Coup expressions are created by RenameInsertions as shorthands for
FeynArts' clumsy G[...][...][...][...] expressions."

RelCF::usage = "RelCF[sym] denotes a relative combinatoric factor in an
InsertionSum expression. RelCF expressions are generated by RenameInsertions."

Sym::usage = "Sym[sym] denotes a dummy symbol in an InsertionSum expression.
Sym expressions are generated by RenameInsertions."


Begin["`Private`"]

(* hack FeynArts`Initialize`ReadModelFile *)

Clear[FeynArts`Initialize`ReadModelFile]

FeynArts`Initialize`ReadModelFile[a___][b___] := Block[
  {$ContextPath=Join[{"HEPMath`FeynArts`Conflict`"},
     DeleteCases[$ContextPath,
       "HEPMath`Tensors`"|"HEPMath`Lorentz`"|"HEPMath`Color`"|
       "HEPMath`FeynArts`"|"HEPMath`PaVe`"]],
   $Context="HEPMath`FeynArts`Model`"},
  oldReadModelFile[a][b]
]

(* add kinematic objects to HEPMath`FeynArts`Model context *)

HEPMath`FeynArts`Model`ChiralityProjector;
HEPMath`FeynArts`Model`DiracSpinor;
HEPMath`FeynArts`Model`DiracMatrix;
HEPMath`FeynArts`Model`DiracSlash;
HEPMath`FeynArts`Model`MetricTensor;
HEPMath`FeynArts`Model`FourVector;
HEPMath`FeynArts`Model`SUNT;
HEPMath`FeynArts`Model`SUNF;
HEPMath`FeynArts`Model`ScalarProduct;
HEPMath`FeynArts`Model`Colour;
HEPMath`FeynArts`Model`Gluon;



(* declare FourMomentum expressions as Lorentz vectors *)

DeclareLorentzVectors[_FourMomentum]


(* ConvertFeynAmp *)

ConvertFeynAmp[FeynAmpList[opt__][amps___]] := Block[
  {proc, min, mout, iin, iout, cfg},

  proc = Process /. {opt};

  min = #[[3]]& /@ First[proc];
  min = MapIndexed[(Mass[Incoming, First[#2]] -> #1)&, min];
  mout = #[[3]]& /@ Last[proc];
  mout = MapIndexed[(Mass[Outgoing, First[#2]] -> #1)&, mout];

  iin =  #[[1]]& /@ First[proc];
  iout = #[[1]]& /@ Last[proc];
  {iin, iout} = {iin, iout} /. (- f:(_F|_S|_V|_U)) -> f;
  iin = If[Length[#] > 1, Cases[#[[2]], _Index], {}]& /@ iin;
  iout = If[Length[#] > 1, Cases[#[[2]], _Index], {}]& /@ iout;
  iin = MapIndexed[(Indices[Incoming, First[#2]] -> #1)&, iin];
  iout = MapIndexed[(Indices[Outgoing, First[#2]] -> #1)&, iout];
  
  cfg = Join[min,mout,iin,iout];
  {cfg, Plus@@(parseAmp[cfg] /@ {amps})}
]

pol[cfg_, io_, n_] := If[FreeQ[cfg, Mass[io, n]],
  Message[ConvertFeynAmp::upol, io, n],
  If[(Mass[io, n] /. cfg) === 0, Pol, MPol]]


insertGI[HoldPattern[Plus[aa__]]] := Plus @@ (insertGI /@ {aa})

insertGI[g_ /; !FreeQ[g,
  _DiracSlash|_DiracMatrix|_ChiralityProjector|_DiracSpinor]] := g

insertGI[m_] := m GI

parseAmp[cfg_][FeynAmp[_,_,ex_,ins_]] := Block[{res={ex,ins}},
  res = res /. NonCommutative[aa__] :> NonCommutative @@ (insertGI /@ {aa});
  res = res //. {
    HEPMath`FeynArts`Conflict`Lorentz -> HEPMath`Lorentz`Lorentz,
    MetricTensor -> Eta,
    DiracMatrix[mu_] -> Ga[mu],
    MatrixTrace[aa__] :> DiracTr[Dot[aa]],
    DiracSlash[p_] -> Gs[p],
    ChiralityProjector[-1] -> PL,
    ChiralityProjector[1] -> PR,
    DiracSpinor[FourMomentum[Incoming,n_],_] ->
      USp[FourMomentum[Incoming,n]][Polarization[Incoming,n]],
    DiracSpinor[-FourMomentum[Incoming,n_],_] ->
      Bar[VSp[FourMomentum[Incoming,n]][Polarization[Incoming,n]]],
    DiracSpinor[FourMomentum[Outgoing,n_],_] ->
      Bar[USp[FourMomentum[Outgoing,n]]][Polarization[Outgoing,n]],
    DiracSpinor[-FourMomentum[Outgoing,n_],_] ->
      VSp[FourMomentum[Outgoing,n]][Polarization[Outgoing,n]],
    FourVector[p_, mu_] -> p[mu],
    Conjugate[PolarizationVector][_, p:FourMomentum[io_, n_], mu_] :>
      CC[pol[cfg, io, n][p]][Polarization[io, n], mu],
    PolarizationVector[_, p:FourMomentum[io_, n_], mu_] :>
      pol[cfg, io, n][p][Polarization[io, n], mu],
    SUNT[a_, i_, j_] -> ColorT[a, i, j],
    SUNF[a_, b_, c_, d_] :> Block[{e=Unique[]}, ColorF[a, b, e] ColorF[e, c, d]],
    SUNF[a_, b_, c_] -> ColorF[a, b, c],
    IndexDelta[Index[Colour, i_], Index[Colour, j_]] ->
      ColorDelta[Index[Colour, i], Index[Colour, j]],
    IndexDelta[Index[Gluon, i_], Index[Gluon, j_]] ->
      GluonDelta[Index[Gluon, i], Index[Gluon, j]],
    ScalarProduct[p_, q_] :> SDot[p, q],
    PropagatorDenominator -> Den,
    _SumOver -> 1
  };
  res = res /. NonCommutative -> Dot;
  res = res /. FermionChain -> Dot;
  res = res /. FeynAmpDenominator -> Times;
  res = res /. Conjugate -> CC;
  If[ins==={}, First[res], InsertionSum@@res]
]

parseAmp[cfg_][FeynAmp[gid_,int_,ex_]] := parseAmp[cfg][FeynAmp[gid,int,ex,{}]]


(* RenameInsertions *)

abbrev[_Mass] := Mass[Unique[]]

abbrev[G[_][_][__][_]] := Coup[Unique[]]

abbrev[RelativeCF] := RelCF[Unique[]]

abbrev[_] := Sym[Unique[]]

renamingRules[_List] = {}

renamingRules[subexl_List -> Insertions[_][ins__]] := 
  Flatten[{MapThread[Rule, {subexl, abbrev /@ subexl}],
           renamingRules /@ {ins}}]

renameInsertions[InsertionSum[ex_, ins_]] :=
  InsertionSum[ex, ins] /. renamingRules[ins]

RenameInsertions[ex_] := ex /. i_InsertionSum :> renameInsertions[i]


(* ApplyInsertions *)

getInsertionRules[subexl_List][ins_List] := {MapThread[Rule, {subexl, ins}]}

getInsertionRules[subexl_List][subsubexl_ -> ins_] :=
  Join[getInsertionRules[subexl][subsubexl], #]& /@
  getInsertionRules[][subsubexl->ins]

getInsertionRules[][subexl_ -> Insertions[_][ins__]] :=
  Flatten[getInsertionRules[subexl] /@ {ins}, 1]

applyInsertions[InsertionSum[ex_, ins_]] :=
  Plus @@ ((ex //. #)& /@ getInsertionRules[][ins])

ApplyInsertions[ex_] := ex /. i_InsertionSum :> applyInsertions[i]


End[]

EndPackage[]


$ContextPath = Join[{"FeynArts`","HEPMath`FeynArts`Model`"},$ContextPath]
