BeginPackage["HEPMath`Kinematics`", {"HEPMath`Tensors`", "HEPMath`Lorentz`"}]

KineticLambda::usage = "KineticLambda[m1sq, m2sq, m3sq] returns the kinetic
function m1sq^2 + m2sq^2 + m3sq^2 - 2 (m1sq m2sq + m2sq m3sq + m3sq m1sq)."

TwoBodyDecay::usage = "TwoBodyDecay[q, m1, m2, theta, phi] returns a list {p1,
p2} where p1 and p2 are the four-momenta of the decay products of the two-body
decay of a particle with four-momentum q into two particles with masses m1 and
m2. The angles theta and phi are polar angles in the decaying particle's rest
frame. The momentum q should be a list with four real numbers and the returned
momenta p1 and p2 will also be lists with four real numbers."

TwoBodyDecay::err = "Cannot decay four momentum `1` into masses `2` and `3`"

Picobarn::usage = "Factor to convert GeV^(-2) into picobarn.
1 pb = Picobarn GeV^(-2)"


Begin["`Protected`"]

cTwoBodyDecay::usage = "compiled version of TwoBodyDecay"

cPol::usage = "compiled version of Pol"

cMPol::usage = "compiled version of MPol"

End[]


Begin["`Private`"]

Picobarn = 1.0/(0.389379338*10^9)

AppendTo[$ContextPath, "HEPMath`Kinematics`Protected`"]

toPolar = Compile[{{p, _Real, 1}},
  Block[{px, py, pz, r, ex, ey, ez, theta, phi},  
    px = p[[2]]; py = p[[3]]; pz = p[[4]];
    r = Re[Sqrt[px^2+py^2+pz^2]];
    ex = 0.0; ey = 0.0; ez = 1.0;
    If[r =!= 0., ex = px/r; ey = py/r; ez = pz/r];
    theta = ArcCos[ez];
    r = Re[Sqrt[ex^2+ey^2]];
    If[r == 0.,
       (* If[ez > 0, phi = 0.0, phi = N[Pi]], *)
       phi = 0.0,
       phi = ArcTan[ex, ey]
    ];
    {theta, phi}
  ],
  CompilationOptions->{"ExpressionOptimization"->True,
                       "InlineCompiledFunctions"->True,
                       "InlineExternalDefinitions"->True}
]

rotateVector = Compile[
  {{tp, _Real, 1},{p, _Real, 1}},
  Block[
    {p0 = p[[1]], px = p[[2]], py = p[[3]], pz = p[[4]],
     ct = Cos[tp[[1]]], st=Sin[tp[[1]]], cp=Cos[tp[[2]]], sp=Sin[tp[[2]]]},
    {p0, cp ct px - py sp + cp pz st, cp py + ct px sp + pz sp st, ct pz - px st}
  ],
  CompilationOptions->{"ExpressionOptimization"->True,
                       "InlineCompiledFunctions"->True,
                       "InlineExternalDefinitions"->True}
]

polarVector = Compile[
  {{m, _Real}, {p, _Real}, {theta, _Real}, {phi, _Real}},
  Block[{ct = Cos[theta], st=Sin[theta], cp=Cos[phi], sp=Sin[phi]},
    {Sqrt[m^2+p^2], st cp p, st sp p, ct p}
  ],
  CompilationOptions->{"ExpressionOptimization"->True,
                       "InlineCompiledFunctions"->True,
                       "InlineExternalDefinitions"->True}
]

boostVector = Compile[{{q, _Real, 1}, {p, _Real, 1}},
  Block[
    {q0=q[[1]], qx=q[[2]], qy=q[[3]], qz=q[[4]],
     p0=p[[1]], px=p[[2]], py=p[[3]], pz=p[[4]],
     msq, m=0., ch=0., sh=0., ang={}},

    msq = q0^2 - qx^2 - qy^2 - qz^2;
    If[msq <= 0, {},
      m = Sqrt[msq];
      ch = q0/m;
      sh = Sqrt[qx^2 + qy^2 + qz^2]/m;
      ang = toPolar[q];
      rotateVector[ang, {ch p0 + sh pz, px, py, sh p0 + ch pz}]
    ]
  ],
  CompilationOptions->{"ExpressionOptimization"->True,
                       "InlineCompiledFunctions"->True,
                       "InlineExternalDefinitions"->True}
]

(* getBoost[{q0_, qx_, qy_, qz_}] := Block[ *)
(*   {msq, m, gm, gm1, btx, bty, btz, btsq}, *)
(*   msq = q0^2 - qx^2 - qy^2 - qz^2; *)
(*   If[msq <= 0, Abort[]]; *)
(*   m = Sqrt[msq]; *)
(*   gm = q0/m; *)
(*   gm1 = gm - 1.0; *)
(*   btx = -qx/(m gm); bty = -qy/(m gm); btz = -qz/(m gm); *)
(*   btsq = btx^2 + bty^2 + btz^2; *)
(*   {{gm, -gm btx, -gm bty, -gm btz}, *)
(*    {-gm btx, (btsq + gm1 btx^2)/btsq, gm1 btx bty/btsq, gm1 btx btz/btsq}, *)
(*    {-gm bty, gm1 bty btx/btsq, (btsq + gm1 bty^2)/btsq, gm1 bty btz/btsq}, *)
(*    {-gm btz, gm1 btz btx/btsq, gm1 btz bty/btsq, (btsq + gm1 btz^2)/btsq}} *)
(* ] *)

cTwoBodyDecay = Compile[
  {{q, _Real, 1}, {m1, _Real}, {m2, _Real}, {theta, _Real}, {phi, _Real}},
  Block[{q0=q[[1]], qx=q[[2]], qy=q[[3]], qz=q[[4]], m0sq, m0, lam, p, p1, p2},
    m0sq = q0^2 - qx^2 - qy^2 - qz^2;
    If[m0sq <= 0, {{},{}},
      m0 = Sqrt[m0sq];
      lam = m0sq^2 + m1^4 + m2^4 - 2. (m0sq m1^2 + m1^2 m2^2 + m2^2 m0sq);
      If[lam <= 0, {{},{}},
        p = Sqrt[lam]/(2 m0);
        p1 = polarVector[m1, p, theta, phi];
        p2 = polarVector[m2, -p, theta, phi];
        {boostVector[q,p1], boostVector[q,p2]}
      ]
    ]
  ],
  CompilationOptions->{"ExpressionOptimization"->True,
                       "InlineCompiledFunctions"->True,
                       "InlineExternalDefinitions"->True}
]

cPol = Compile[{{k, _Real, 1}, {sg, _Integer}},
  rotateVector[toPolar[k], {0.0, 1.0/Sqrt[2.0], N[sg] I/Sqrt[2.0], 0.0}],
  CompilationOptions->{"ExpressionOptimization"->True,
                       "InlineCompiledFunctions"->True,
                       "InlineExternalDefinitions"->True}
]

cMPol = Compile[{{k, _Real, 1}, {sg, _Integer}},
  Block[{k0, kx, ky, kz, p, msq, m},
    If[sg == 0,
      k0 = k[[1]]; kx = k[[2]]; ky = k[[3]]; kz = k[[4]];
      p = Sqrt[kx^2 + ky^2 + kz^2];
      msq = k0^2 - kx^2 - ky^2 - kz^2;
      If[msq <= 0, {},
        m = Sqrt[msq];
        {p/m, k0 kx/(p m), k0 ky/(p m), k0 kz/(p m)}
      ]
      ,
      rotateVector[toPolar[k], {0.0, 1.0/Sqrt[2.0], N[sg] I/Sqrt[2.0], 0.0}]
    ]
  ],
  CompilationOptions->{"ExpressionOptimization"->True,
                       "InlineCompiledFunctions"->True,
                       "InlineExternalDefinitions"->True}
]

KineticLambda[x_, y_, z_] = x^2 + y^2 + z^2 - 2(x y + y z + z x)

nrealQ[x_] := Element[x, Reals] === True

TwoBodyDecay[q:{__?nrealQ}, m1_?nrealQ, m2_?nrealQ,
             theta_?nrealQ, phi_?nrealQ] := Block[{res},
  res = cTwoBodyDecay[N /@ q, N[m1], N[m2], N[theta], N[phi]];
  If[res === {}, Message[TwoBodyDecay::err, q, m1, m2]; Abort[]];
  res
]

Pol[k:{_?nrealQ, _?nrealQ, _?nrealQ, _?nrealQ}][sg:((-1)|1)] :=
  cPol[N /@ k, sg]

CC[Pol[k:{_?nrealQ, _?nrealQ, _?nrealQ, _?nrealQ}]][sg:((-1)|1)] :=
  Conjugate[Pol[k][sg]]

MPol[k:{_?nrealQ, _?nrealQ, _?nrealQ, _?nrealQ}][sg:((-1)|0|1)] :=
  cMPol[N /@ k, sg]

CC[MPol[k:{_?nrealQ, _?nrealQ, _?nrealQ, _?nrealQ}]][sg:((-1)|0|1)] :=
  Conjugate[MPol[k][sg]]

$ContextPath = DeleteCases[$ContextPath, "HEPMath`Kinematics`Protected`"]

End[]

EndPackage[]

