(* Load HEPMath. *)
Needs["HEPMath`"]

(* Load FeynArts interface. *)
Needs["HEPMath`FeynArts`"]

(* Load compilation module. *)
Needs["HEPMath`Compile`"]

Needs["HEPMath`LHAPDF`"]

(* Tell FeynArts to shut up. *)
$FAVerbose = 0

(* Use anti-commuting gamma_5. *)
SetOptions[CalcDiracTraces, AntiCommutingG5 -> True]

(* Initialize SM model *)
InitializeModel["SMQCD"]


(* Name of the process. We'll use this as prefix for generated files. *)
name = "H-gg-SM"

(* FeynArts process definition: H -> g g. *)
process = {S[1]} -> {V[5], V[5]}

(* Set options for field insertions *)
SetOptions[InsertFields, Model -> "SMQCD", Restrictions -> NoLightFHCoupling,
  LastSelections->{F[3,{3,___}]}, InsertionLevel->{Particles}]

(* Set options for drawing diagrams. *)
SetOptions[Paint, PaintLevel -> {Particles}, ColumnsXRows -> {3, 4}]

(* Generate topologies. *)
tops = CreateTopologies[1, 1 -> 2, ExcludeTopologies->{Tadpoles, WFCorrections}];
(* Insert fields. *)
ins = InsertFields[tops, process];
(* Draw diagrams and export to ps *)
Export[name<>"-diagrams.ps", Paint[ins]];

(* Insert Feynman rules. *)
amp = CreateFeynAmp[ins];
(* Convert to HEPMath notation *)
{cfg, amp} = ConvertFeynAmp[amp];


(* Now we have to compute the amplitude 'amp' *)

(* Declare various parameters as real *)
DeclareReal[{muR,  (* Renormalization scale *)
             GS,   (* Strong coupling constant *)
             EL,   (* QED coupling constant *)
             SW,   (* Sine of the Weinberg angle *)
             CW,   (* Cosine of the Weinberg angle *)
             MW,   (* W mass *)
             MZ,   (* Z mass *)
             MH}]  (* Higgs mass *)

(* Declare Lorentz vectors *)
DeclareLorentzVectors[_k, _p, _l];
DeclareComplexLorentzVectors[_ef]

(* Shorten notations in the amplitude generated by FeynArts a bit *)
amp = amp /. {FourMomentum[Incoming, n_] -> k[n],   (* Incoming momenta *)
              FourMomentum[Outgoing, n_] -> p[n],   (* Outgoing momenta *)
              FourMomentum[Internal, n_] -> l[n],   (* Loop momentum *)
              Index[Gluon, n_]           -> a[n],   (* Gluon indices *)
              Index[Colour, n_]          -> i[n],   (* Color indices *)
              Index[Lorentz, n_]         -> mu[n]}  (* Lorentz indices *)

(* Introduce a simpler notation for polarization vectors *)
amp = amp /. {CC[Pol[p[i_]]][_, mu_] -> ef[i][mu]}

(* Use momentum conservation *)
amp = amp /. p[1] -> k[1] - p[2]

(* Substitutions for scalar products of external momenta and polarization
   vectors. We'll have to do this several times so we define a function *)
kinSimplify[ex_] := Block[{res},
  res = HEPExpand[ex, p|k|l];
  res = LorentzContract[res];
  res = res /. {
    SDot[Pol[p_][_], p] -> 0,
    SDot[CC[Pol[p_]][_], p] -> 0,
    Sqr[k[1]] -> MH^2,
    Sqr[p[2]] -> 0,
    SDot[k[1], p[2]] -> MH^2/2,

    SDot[p[2], ef[2]] -> 0,
    SDot[p[2], ef[1]] -> 0,
    SDot[k[1], ef[1]] -> 0,
    SDot[k[1], ef[2]] -> 0
  }
]

(* Simplify the color structures *)
amp = ColorReduce[amp]

(* Calculate the Dirac traces. To speed things up we first collect them. *)
amp = HEPCollect[DiracExpand[amp], DiracTr];
amp = CalcDiracTraces[amp]

(* Substitute kinematics *)
amp = kinSimplify[amp]

(* Do the reduction to tensor integrals *)
amp = PaVeIntegrate[amp, l[1], muR]

(* Simplify the arguments of the PaVe expressions *)
amp = ApplyToPaVeArgs[kinSimplify, amp]

(* Substitute scalar products (again) *)
amp = kinSimplify[amp]

(* Check if the amplitude is UV finite *)
amp = amp /. Dim -> 4 - 2/Div;
amp = Collect[amp, Div];
If[!FreeQ[amp, Div],
  WriteString[$Output, "ERROR: Amplitude is not UV finite.\n"];
  Exit[1];
]

(* Split the amplitude into color structures and color amplitudes.
   There is actually only one color structure in this problem, but I show the
   general procedure here in case you want to calculate more complicated
   processes. *)
{colorstructs, coloramps} =
  HEPCollectListed[amp, ColorT|ColorF|ColorDelta|ColAdDelta]

(* Contracting all possible combinations of color structures gives you
   the color matrix *)
colormat = Outer[ColorReduce[CC[#1] #2]&, colorstructs, colorstructs]


(* So much for the symbolic part of the calculation. Now we do the numerics.
   We will use HEPCompile to compile a function which takes alpha_s, the Higgs
   mass and the top mass as arguments and returns the partial H->gg width. *)

(* First we specify values for the parameters we don't want to change. *)
AlphaEM = 1.0/128.0;
EL = N[Sqrt[4 Pi*AlphaEM]];
MW = 80.399;
MZ = 91.1876;
CW = MW/MZ;
SW = Sqrt[1.0 - CW^2];

(* Work in four dimensions *)
Dim = 4

(* Set the renormalization scale to the top mass.
   Note that we have not specified a value for the top mass. *)
muR = MT


(* HEPCompile can handle expressions that contain Lorentz vectors, but these
   vectors must be represented by pure symbols. We therefore have to change
   our notation a little bit. *)
DeclareComplexLorentzVectors[k1, p1, p2, ef1, ef2];
coloramps = coloramps /. {k[1] -> k1, p[2] -> p2, ef[1] -> ef1, ef[2] -> ef2}
  

(* Set options for compilation. *)
SetOptions[HEPCompile,
  CompilationOptions->{"ExpressionOptimization"->True}]

(* Compile the partial H->gg decay width as a function of AlphaS, MH and MT.
   The multiplication with the color matrix and the sum over gluon polarisations 
   is done numerically inside the function.

   Long expressions can be substituted in the function body with HEPEvaluate
   or HEPPrepare. Use the latter if the expressions contain PaVe, SDot or Sqr.
   It is your responsibility to assure that all symbols appearing in substituted
   expressions are assigned a value before they appear on the right-hand side
   of an assignment.
 *)
gammaHgg = HEPCompile[
  {{AlphaS, _Real}, {MH, _Real}, {MT, _Real}}, (* Function arguments. *)

  (* Function body. Use Block to create local variables *)
  Block[{k1, p1, p2, ef1, ef2,  (* External momenta and polarisation vectors *)
         GS,                    (* Strong coupling constant *)
         camps,                 (* Color amplitudes *)
         polsum,                (* Squared matrix element summed over
                                   polarizations *)
         psfac},                (* Phase space factor *)

    (* Set incoming momentum vector *)
    k1 = {MH, 0., 0., 0.};
    (* Construct final state momenta with TwoBodyDecay *)
    {p1, p2} = TwoBodyDecay[k1, 0., 0., 0., 0.];
    (* Compute the strong coupling constant from AlphaS *)
    GS = Sqrt[4 Pi*AlphaS];

    (* Loop over polarizations. *)
    polsum = 0.;
    Do[
      (* Set the polarization vectors *)
      ef1 = CC[Pol[p1]][lm1];
      ef2 = CC[Pol[p2]][lm2];
      (* Store the color form factors in a local variable.
         Note the use of HEPPrepare. *)
      camps = HEPPrepare[coloramps];
      (* Contract color form factors with color matrix.
         Not the use of HEPEvaluate to substitute the color matrix. *)
      polsum = polsum + Re[Conjugate[camps].HEPEvaluate[colormat].camps];
      ,
      {lm1, {-1,+1}}, {lm2, {-1,+1}}
    ];

    (* Compute phase space factor *)
    psfac = Sqrt[KineticLambda[MH^2, 0, 0]]/(8 Pi MH^2);

    (* Return partial decay width. Note the symmetry factor 1/2. *)
    (1/2)*polsum*psfac/(2 MH)
  ]
]


(* Now call the compiled function and print the result *)

AlphaS = 0.118;
MT = 172.0;
MH = 125.0;
gamHgg = gammaHgg[AlphaS, MH, MT];
WriteString[$Output, "Gamma(H->gg) = "<>ToString[gamHgg]<>" GeV\n"]


(* Compare with literature formula *)
gammaHggLit[alphas_, mh_, mt_] := Block[{t, ft, AQ, GF},
  GF = Pi AlphaEM/(Sqrt[2] SW^2 MW^2);

  (* hep-ph/9504378, Eq. 2 *)
  t = mh^2/(4 mt^2);

  (* hep-ph/9504378, Eq. 4 *)
  ft = If[t <= 1,
    ArcSin[Sqrt[t]]^2,
    -1/4 (Log[(1 + Sqrt[1-1/t])/(1 - Sqrt[1-1/t])] - I Pi)^2];

  (* hep-ph/9504378, Eq. 3 *)
  AQ = 2 (t + (t-1) ft)/t^2;

  (* hep-ph/9504378, Eq. 21 *)
  GF alphas^2 mh^3/(36 Sqrt[2] Pi^3) Abs[3/4 AQ]^2
];
gamHggLit = gammaHggLit[AlphaS, MH, MT];
If[Abs[gamHgg - gamHggLit]/(Abs[gamHgg] + Abs[gamHggLit]) > 10^(-10),
  WriteString[$Output, "ERROR: Disagreement with literature result: "<>
  ToString[gamHggLit]<>" GeV\n"]]


(* Put the function into a Python module.

   If the compilation worked you can go into the 'H-gg-SM.build' directory and
   run

   python
   >>> from gammaHgg import *
   >>> gammaHgg(0.118, 125.0, 172.0)   
   0.00022152990766551457

   See hepcompile.m example for more explanations
*)

builddir = name<>".build";
libname = "gammaHgg";

If[!DirectoryQ[builddir], CreateDirectory[builddir]]
modules = HEPCodeGenerate[{gammaHgg}, {"gammaHgg"}, libname,
  TargetDirectory->builddir];

Needs["CCompilerDriver`"];

CreateLibrary[modules, "_"<>libname,
  (* "ShellCommandFunction"->Print, *)  (* Uncomment this to see the
                                           compile command *)
  "TargetDirectory" -> builddir,
  "Debug"->True, "Libraries"->{"WolframRTL", "dl"}
];

DeleteDirectory[#, DeleteContents->True]& /@
  FileNames[FileNameJoin[{builddir, "Working*"}]]

