__all__ = [
    'c_mint', 'c_mreal', 'c_mcomplex', 'c_MTensor', 
    'c_mint_p', 'c_mreal_p', 'c_mcomplex_p', 'c_MTensor_p',
    'mint_t', 'mreal_t', 'mcomplex_t',
    'mType', 'npType',
    'getNumpyArray', 'mtensor', 'mtensor_new', 'mtensor_free'
]

import math
import cmath
import numpy as np
import ctypes as ct


# ctypes for Mathematica interface

c_mint = ct.c_long
c_mreal = ct.c_double

c_mint_p = ct.POINTER(c_mint)
c_mreal_p = ct.POINTER(c_mreal)

class c_mcomplex(ct.Structure):
    _fields_ = [('real', ct.c_double), ('imag', ct.c_double)]

    def __init__(self, *args):
        if len(args) == 1:
            ct.Structure.__init__(self, args[0].real, args[0].imag)
        else:
            ct.Structure.__init__(self, *args)

    @property
    def value(self):
        return complex(self.real, self.imag)

c_mcomplex_p = ct.POINTER(c_mcomplex)

class c_TENSOR_PROPERTY_DATA_STRUCT(ct.Structure):
    _fields_ = [('prec',  c_mreal),
                ('dims',  ct.POINTER(c_mint)),
                ('rank',  c_mint),
                ('type',  ct.c_int),
                ('flags', ct.c_uint)]

class c_M_TENSOR_STRUCT(ct.Structure):
    _fields_ = [('data', ct.c_void_p),
                ('properties', ct.POINTER(c_TENSOR_PROPERTY_DATA_STRUCT)),
                ('flattened_length', c_mint)]

c_MTensor = ct.POINTER(c_M_TENSOR_STRUCT)

c_MTensor_p = ct.POINTER(c_MTensor)


# Mathematica's int, real and complex types
# PROBABLY PLATFORM DEPENDENT
mint_t     = 'int64'
mreal_t    = 'float64'
mcomplex_t = 'complex128'

# type dictionaries
mType = {mint_t:2, mreal_t:3, mcomplex_t:4}
npType = {2:np.dtype(mint_t), 3:np.dtype(mreal_t), 4:np.dtype(mcomplex_t)}

# Mathematica's flags for tensor objects. No clue what they mean.
# mTensorFlags = 139
mTensorFlags = 131

# buffer class for reading result tensors
class Buffer:
    def __init__(self, version=3, typestr=None, data=None, shape=None):
        self.__array_interface__ = {
            'version': version,
            'typestr': typestr,
            'data': (data, False),
            'shape': shape
        }

def getNumpyArray(t):
    """
    Extract the data from a c_MTensor object and return it as a numpy array

    """
    properties = t.contents.properties.contents
    rank = properties.rank
    shape = tuple(properties.dims[i] for i in range(rank))
    mtype = properties.type

    pdata = t.contents.data
    dtype = np.dtype(npType[mtype])

    buf = Buffer(typestr=dtype.str, data=pdata, shape=shape)
    a = np.array(buf, copy=True)
    return a
    

def mtensor(a, dtype=None, rank=None):
    """
    Converts a numpy array into a c_MTensor object (for use with ctypes)

    """
    if dtype is None:
        dtype = a.dtype
    else:
        dtype = np.dtype(dtype)
    if dtype.name not in mType:
        raise ValueError('only integer, float and complex arrays are supported')
    a = np.require(a, dtype, ['C', 'A'])
    n = len(a.shape)
    if rank is not None and n != rank:
        raise ValueError('tensor argument has wrong rank')
    prec = c_mreal(-np.inf)
    dims = (c_mint*n)(*(a.shape))
    pdims = ct.cast(ct.pointer(dims), ct.POINTER(c_mint))
    rank = c_mint(n)
    tpe = ct.c_int(mType[dtype.name])
    flags = ct.c_uint(mTensorFlags)
    properties = c_TENSOR_PROPERTY_DATA_STRUCT(prec, pdims, rank, tpe, flags)
    data = ct.cast(a.ctypes.data, ct.c_void_p)
    length = c_mint(reduce(lambda x, y: x*y, a.shape, 1))
    t = c_M_TENSOR_STRUCT(data, ct.pointer(properties), length)
    return a, ct.pointer(t)

def mtensor_new(lib, tpe):
    """
    Allocates a new MTensor object.

    Args:
      lib: ctypes library object to use.
      tpe: Data type of tensor. Should be mint_t, mreal_t or mcomplex_t.

    Returns:
      A c_MTensor object representing the created MTensor structure.

    """
    res = c_MTensor()
    err = lib.HEPMath_MTensor_new(c_mint(mType[tpe]), ct.pointer(res))
    if err != 0:
        raise RuntimeError(
            'allocation of MTensor failed with error code '+str(err))
    return res
    
def mtensor_free(lib, mt):
    """
    Frees the memory of a c_MTensor object.

    Args:
      lib: ctypes library object to use.
      mt: c_MTensor object to free.

    """
    lib.HEPMath_MTensor_free(mt)
